@@ -1,10 +1,9 @@
 /* saslutil.c
  * Rob Siemborski
  * Tim Martin
- * $Id: saslutil.c,v 1.44.2.1 2009/04/27 17:47:17 murch Exp $
  */
 /* 
- * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
+ * Copyright (c) 1998-2016 Carnegie Mellon University.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -22,12 +21,13 @@
  *    endorse or promote products derived from this software without
  *    prior written permission. For permission or any other legal
  *    details, please contact  
- *      Office of Technology Transfer
  *      Carnegie Mellon University
- *      5000 Forbes Avenue
- *      Pittsburgh, PA  15213-3890
- *      (412) 268-4387, fax: (412) 268-7395
- *      tech-transfer@andrew.cmu.edu
+ *      Center for Technology Transfer and Enterprise Creation
+ *      4615 Forbes Avenue
+ *      Suite 302
+ *      Pittsburgh, PA  15213
+ *      (412) 268-7393, fax: (412) 268-7395
+ *      innovation@andrew.cmu.edu
  *
  * 4. Redistributions of any form whatsoever must retain the following
  *    acknowledgment:
@@ -43,6 +43,10 @@
  * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#if defined(WIN32)
+#define _CRT_RAND_S
+#endif
+
 #include <config.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -72,7 +76,8 @@
  * sasl_randseed
  * sasl_rand
  * sasl_churn
-*/
+ * sasl_erasebuffer
+ */
 
 char *encode_table;
 char *decode_table;
@@ -111,29 +116,30 @@ static char index_64[128] = {
  * Returns SASL_OK on success, SASL_BUFOVER if result won't fit
  */
 
-int sasl_encode64(const char *_in, unsigned inlen,
-		  char *_out, unsigned outmax, unsigned *outlen)
+int sasl_encode64(const char *_in,
+		  unsigned inlen,
+		  char *_out,
+		  unsigned outmax,
+		  unsigned *outlen)
 {
     const unsigned char *in = (const unsigned char *)_in;
     unsigned char *out = (unsigned char *)_out;
     unsigned char oval;
-    char *blah;
     unsigned olen;
 
     /* check params */
-    if ((inlen >0) && (in == NULL)) return SASL_BADPARAM;
+    if ((inlen > 0) && (in == NULL)) return SASL_BADPARAM;
     
     /* Will it fit? */
     olen = (inlen + 2) / 3 * 4;
     if (outlen) {
-      *outlen = olen;
+	*outlen = olen;
     }
     if (outmax <= olen) {
-      return SASL_BUFOVER;
+	return SASL_BUFOVER;
     }
 
     /* Do the work... */
-    blah=(char *) out;
     while (inlen >= 3) {
       /* user provided max buffer size; make sure we don't go over it */
         *out++ = basis_64[in[0] >> 2];
@@ -222,6 +228,10 @@ int sasl_decode64(const char *in,
         }
     }
 
+    *out = '\0'; /* NUL terminate the output string */
+
+    if (outlen) *outlen = len;
+
     if (inlen != 0) {
         if (saw_equal) {
             /* Unless there is CRLF at the end? */
@@ -231,10 +241,6 @@ int sasl_decode64(const char *in,
         }
     }
 
-    *out = '\0'; /* NUL terminate the output string */
-
-    if (outlen) *outlen = len;
-
     return SASL_OK;
 }
 
@@ -274,9 +280,9 @@ int sasl_mkchal(sasl_conn_t *conn,
   time(&now);
 
   if (hostflag && conn->serverFQDN)
-    snprintf(buf,maxlen, "<%lu.%lu@%s>", randnum, now, conn->serverFQDN);
+    snprintf(buf,maxlen, "<%lu.%lu@%s>", randnum, (unsigned long)now, conn->serverFQDN); /* don't care much about time 32bit overlap */
   else
-    snprintf(buf,maxlen, "<%lu.%lu>", randnum, now);
+    snprintf(buf,maxlen, "<%lu.%lu>", randnum, (unsigned long)now);
 
   return (int) strlen(buf);
 }
@@ -295,7 +301,7 @@ int sasl_utf8verify(const char *str, unsigned len)
     if (seqlen == 1) return SASL_BADPROT; /* this shouldn't happen here */
     if (seqlen > 6) return SASL_BADPROT; /* illegal */
     while (--seqlen)
-      if ((str[++i] & 0xC0) != 0xF0) return SASL_BADPROT; /* needed a 10 octet */
+      if ((str[++i] & 0xC0) != 0x80) return SASL_BADPROT; /* needed a 10 octet */
   }
   return SASL_OK;
 }      
@@ -422,6 +428,11 @@ static void randinit(sasl_rand_t *rpool)
 	srandom(*foo);
     }
 #endif /* HAVE_JRAND48 */
+#elif defined(WIN32)
+    {
+	unsigned int *foo = (unsigned int *)rpool->pool;
+	srand(*foo);
+    }
 #endif /* WIN32 */
     }
 
@@ -430,24 +441,35 @@ static void randinit(sasl_rand_t *rpool)
 void sasl_rand (sasl_rand_t *rpool, char *buf, unsigned len)
 {
     unsigned int lup;
+#if defined(WIN32) && !defined(__MINGW32__)
+    unsigned int randomValue;
+#endif
+
     /* check params */
     if (!rpool || !buf) return;
     
     /* init if necessary */
     randinit(rpool);
- 
-#if (defined(WIN32)||defined(macintosh))
-    for (lup=0;lup<len;lup++)
+
+    for (lup = 0; lup < len; lup++) {
+#if defined(__MINGW32__)
 	buf[lup] = (char) (rand() >> 8);
-#else /* WIN32 */
+#elif defined(WIN32)
+	if (rand_s(&randomValue) != 0) {
+	    randomValue = rand();
+	}
+
+	buf[lup] = (char) (randomValue >> 8);
+#elif defined(macintosh)
+	buf[lup] = (char) (rand() >> 8);
+#else /* !WIN32 && !macintosh */
 #ifdef HAVE_JRAND48
-    for (lup=0; lup<len; lup++)
 	buf[lup] = (char) (jrand48(rpool->pool) >> 8);
 #else
-    for (lup=0;lup<len;lup++)
 	buf[lup] = (char) (random() >> 8);
 #endif /* HAVE_JRAND48 */
 #endif /* WIN32 */
+    }
 }
 
 /* this function is just a bad idea all around, since we're not trying to
@@ -470,7 +492,114 @@ void sasl_erasebuffer(char *buf, unsigned len) {
     memset(buf, 0, len);
 }
 
+/* Lowercase string in place */
+char *sasl_strlower (
+  char *val
+)
+{
+    int i;
+
+    if (val == NULL) {
+	return (NULL);
+    }
+
+/* don't use tolower(), as it is locale dependent */
+
+    for (i = 0; val[i] != '\0'; i++) {
+	if (val[i] >= 'A' && val[i] <= 'Z') {
+	    val[i] = val[i] - 'A' + 'a';
+	}
+    }
+
+    return (val);
+}
+
+/* A version of gethostname that tries hard to return a FQDN */
+int get_fqhostname(
+  char *name,  
+  int namelen,
+  int abort_if_no_fqdn
+)
+{
+    int return_value;
+    struct addrinfo hints;
+    struct addrinfo *result;
+
+    return_value = gethostname (name, namelen);
+    if (return_value != 0) {
+	return (return_value);
+    }
+
+    if (strchr (name, '.') != NULL) {
+	goto LOWERCASE;
+    }
+
+/* gethostname hasn't returned a FQDN, we have to canonify it ourselves */
+    hints.ai_family = PF_UNSPEC;
+    hints.ai_flags = AI_CANONNAME;
+    hints.ai_socktype = SOCK_STREAM;	/* TCP only */
+/* A value of zero for ai_protocol indicates the caller will accept any protocol. or IPPROTO_TCP? */
+    hints.ai_protocol = 0;  /* 0 or IPPROTO_xxx for IPv4 and IPv6 */
+    hints.ai_addrlen = 0;
+    hints.ai_canonname = NULL;
+    hints.ai_addr = NULL;
+    hints.ai_next = NULL;
+
+    if (getaddrinfo(name,
+		  NULL,		/* don't care abour service/port */
+		  &hints,
+		  &result) != 0) {
+        if (abort_if_no_fqdn) {
+	    /* errno on Unix, WSASetLastError on Windows are already done by the function */
+	    return (-1);
+	} else {
+	    goto LOWERCASE;
+	}
+    }
+
+    if (result == NULL || result->ai_canonname == NULL) {
+	freeaddrinfo (result);
+        if (abort_if_no_fqdn) {
 #ifdef WIN32
+	    WSASetLastError (WSANO_DATA);
+#elif defined(ENODATA)
+	    errno = ENODATA;
+#elif defined(EADDRNOTAVAIL)
+	    errno = EADDRNOTAVAIL;
+#endif
+	    return (-1);
+	} else {
+	    goto LOWERCASE;
+	}
+    }
+
+    if (strchr (result->ai_canonname, '.') == NULL) {
+	freeaddrinfo (result);
+        if (abort_if_no_fqdn) {
+#ifdef WIN32
+	    WSASetLastError (WSANO_DATA);
+#elif defined(ENODATA)
+	    errno = ENODATA;
+#elif defined(EADDRNOTAVAIL)
+	    errno = EADDRNOTAVAIL;
+#endif
+	    return (-1);
+	} else {
+	    goto LOWERCASE;
+	}
+    }
+
+
+/* Do we need to check for buffer overflow and set errno? */
+    strncpy (name, result->ai_canonname, namelen);
+    freeaddrinfo (result);
+
+LOWERCASE:
+    sasl_strlower (name);
+    return (0);
+}
+
+#if defined(WIN32) && !defined(__MINGW64_VERSION_MAJOR)
 /***************************************************************************** 
  * 
  *  MODULE NAME : GETOPT.C 
@@ -657,13 +786,13 @@ getpass(prompt)
 const char *prompt;
 {
 	register char *p;
-	register c;
+	register int c;
 	static char pbuf[PASSWORD_MAX];
 
 	fprintf(stderr, "%s", prompt); (void) fflush(stderr);
 	for (p=pbuf; (c = _getch())!=13 && c!=EOF;) {
 		if (p < &pbuf[sizeof(pbuf)-1])
-			*p++ = c;
+			*p++ = (char) c;
 	}
 	*p = '\0';
 	fprintf(stderr, "\n"); (void) fflush(stderr);
