@@ -155,6 +155,9 @@ int main(int argc, char **argv) {
 	char            *auth_mech_name = NULL;
 	size_t		pid_file_size;
 
+	/* XXX  force openlog() before any of our mechs try syslog() */
+	logger(L_INFO, L_FUNC, "starting %s", argv[0]);
+
 	SET_AUTH_PARAMETERS(argc, argv);
 
 	g_argc = argc;
@@ -248,6 +251,14 @@ int main(int argc, char **argv) {
 		exit(1);
 	}
 
+	/* Create our working directory */
+	if (mkdir(run_path, 0755) == -1 && errno != EEXIST) {
+		logger(L_ERR, L_FUNC, "can not mkdir: %s", run_path);
+		logger(L_ERR, L_FUNC, "Check to make sure the parent directory exists and is");
+		logger(L_ERR, L_FUNC, "writeable by the user this process runs as.");
+		exit(1);
+	}
+
 	set_auth_mech(auth_mech_name);
 
 	if (flags & VERBOSE)  {
@@ -276,7 +287,7 @@ int main(int argc, char **argv) {
 		exit(1);
 	}
 
-	umask(077);
+	umask(0077);
 
 	pid_file_size = strlen(run_path) + sizeof(PID_FILE_LOCK) + 1;
 	if ((pid_file_lock = malloc(pid_file_size)) == NULL) {
@@ -287,7 +298,7 @@ int main(int argc, char **argv) {
 	strlcpy(pid_file_lock, run_path, pid_file_size);
 	strlcat(pid_file_lock, PID_FILE_LOCK, pid_file_size);
 
-	if ((pid_file_lock_fd = open(pid_file_lock, O_CREAT|O_TRUNC|O_RDWR, 644)) < 0) {
+	if ((pid_file_lock_fd = open(pid_file_lock, O_CREAT|O_TRUNC|O_RDWR, 0644)) < 0) {
 		rc = errno;
 		logger(L_ERR, L_FUNC, "could not open pid lock file: %s", pid_file_lock);
 		logger(L_ERR, L_FUNC, "open: %s", strerror(rc));
@@ -370,7 +381,7 @@ int main(int argc, char **argv) {
 char *do_auth(const char *_login, const char *password, const char *service, const char *realm) {
 
 	struct cache_result	lkup_result;
-	char			*response;
+	char			*response = NULL;
 	int			cached = 0;
 	char			login_buff[MAX_LOGIN_REALM_LEN];
 	char			*login;
@@ -427,6 +438,7 @@ char *do_auth(const char *_login, const char *password, const char *service, con
 	}
 
 	logger(L_ERR, L_FUNC, "mechanism returned unknown response: %s", auth_mech->name);
+	free(response);
 	response = strdup("NO internal mechanism failure");
 
 	return response;
@@ -620,7 +632,6 @@ void detach_tty() {
     int		null_fd;
     int         exit_result;
     pid_t      	pid;
-    char       	pid_buf[100];
     struct flock	lockinfo;
     
     /**************************************************************
@@ -882,7 +893,6 @@ void handle_sigchld() {
  * Do some final cleanup here.
  **************************************************************/
 void server_exit() {
-	struct flock    lock_st;
 
 	/*********************************************************
 	 * If we're not the master process, don't do anything
